import pandas as pd
import os
from cohesion.calculations import calculate_cohesion_score
from cohesion.utils import create_topic_names_using_tf_idf, extract_topics_labels, detailed_topics_print, \
    input_path_to_df, create_future_dirs, delete_tmp_dirs


def cohesion():
    print("Hello - welcome to cohesion project.")


def cohesion_df(data, docs_percent=1, topics_percent=1, del_tmp=True):
    if type(data) is str:
        print("input is str")
        data = input_path_to_df(data)
    elif type(data) is pd.DataFrame:
        print("input is dataframe")
    else:
        raise TypeError("Bad input type")
    return run_cohesion(data, docs_percent, topics_percent, del_tmp)
#
#
def run_cohesion(df, docs_percent, topics_percent, del_tmp):
    print('----------- Starting main pipeline -------------')
    docs = df['text'].tolist()
    labels = df['label'].tolist()
    create_future_dirs()
    topics = create_topic_names_using_tf_idf(docs, labels)
    topics_names = extract_topics_labels(topics)
    detailed_topics_print(topics_names)
    cohesion_score = calculate_cohesion_score(ground_truth_docs_list=docs,
                                              division_labels=labels, topics_names=topics_names, docs_percent=docs_percent, topics_percent=topics_percent)
    topics_values = [" ".join(topic_words) for topic_words in topics_names]
    if del_tmp:
        delete_tmp_dirs()
    return cohesion_score, topics_values



from cohesion.utils import docs_to_groups, create_input_to_mnli_topics_content, create_topic_id_topic_value
from cohesion.zero_shot_classifier import ZeroShotClassifier
from cohesion.formula import cohesion_calculate


def calculate_cohesion_score(ground_truth_docs_list, division_labels, topics_names, docs_percent, topics_percent):
    # prepare pickles files
    path_to_save = '.\\tmp_coh_exports\\mnli\\doc_topic_mnli_results'
    topic_id_content = '.\\tmp_coh_exports\\topics_map\\topic_id_content.pkl'
    topic_id_value_map = '.\\tmp_coh_exports\\topics_map\\topic_id_topic_value_map.pkl'

    clusters_docs = docs_to_groups(ground_truth_docs_list, division_labels)
    create_input_to_mnli_topics_content(clusters_docs, topic_id_content)
    create_topic_id_topic_value(sorted(clusters_docs.keys()), topics_names, topic_id_value_map)

    # create csv with mnli score
    zero_shot_classifier = ZeroShotClassifier(topic_id_value_map, topic_id_content, division_labels, ground_truth_docs_list)
    zero_shot_classifier.run_classify(division_labels, path_to_save, docs_percent, topics_percent)
    return cohesion_calculate(path=path_to_save + '.csv')


import os
import random
import pandas as pd
from nltk.tokenize import word_tokenize
from sklearn.feature_extraction.text import TfidfVectorizer
import nltk
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
import shutil


nltk.download('stopwords')
nltk.download('punkt')
stop_words = set(stopwords.words('english'))


def tokenizer(text):
    word_tokens = word_tokenize(text)
    bad_words_list = ["http"]
    filtered_sentence = [w.lower() for w in word_tokens if
                         not w.lower() in stop_words and w.isalpha() and len(w) > 2 and w not in bad_words_list]
    return filtered_sentence


def docs_to_groups(docs, clusters):
    clusters_docs = dict()
    for index, doc in enumerate(docs):
        cluster = int(clusters[index])
        docs_list = clusters_docs.get(cluster, list())
        docs_list.append(doc)
        clusters_docs[cluster] = docs_list
    return clusters_docs


def create_input_to_mnli_topics_content(content_dict, path):
    content_list = []
    topics_ind_list = []
    for topic_ind in sorted(content_dict.keys()):
        content_list += content_dict[topic_ind]
        topics_ind_list += [topic_ind] * len(content_dict[topic_ind])
    df_topic_id_content = pd.DataFrame({"topic_ind": topics_ind_list, "content": content_list})
    df_topic_id_content.to_pickle(path, protocol=4)


def create_topic_id_topic_value(topics_ind_list, topic_values, path):
    topic_values_joined = []
    for ten_words_seperated in topic_values:
        topic_values_joined.append(" ".join(ten_words_seperated))
    df_topic_id_value_coherence = pd.DataFrame({"topic_ind": topics_ind_list, "topic_name": topic_values_joined})
    df_topic_id_value_coherence.to_pickle(path, protocol=4)


def tf_idf_top_words(docs: list, words_count=6):
    vectorizer = TfidfVectorizer(use_idf=True, stop_words='english', ngram_range=(1, 1), tokenizer=tokenizer)
    vectors = vectorizer.fit_transform(docs)
    feature_names = vectorizer.get_feature_names()
    dense = vectors.todense()
    denselist = dense.tolist()
    df = pd.DataFrame(denselist, columns=feature_names)
    return list(df.mean(axis=0).sort_values(ascending=False).index)[0:words_count]


def docs_to_clusters(docs_list, docs_label):
    clusters_docs = dict()
    for index, label in enumerate(docs_label):
        doc_list = clusters_docs.get(label, list())
        doc_list.append(docs_list[index])
        clusters_docs[label] = doc_list
    return clusters_docs


def get_division(docs_label):
    clusters_docs = dict()
    for index, label in enumerate(docs_label):
        index_list = clusters_docs.get(label, list())
        index_list.append(index)
        clusters_docs[label] = index_list
    return clusters_docs


def create_topic_name(key, labels_array):
    topic_name = str(key)
    for label in labels_array:
        topic_name += '_' + label
    return topic_name


def create_topic_names_using_tf_idf(docs, labels):
    """
    docs - list of all docs values.
    labels - labels as list.
    returns dict of key:label ind, value: list of docs-related.
    """
    tf_idf_labels = dict()
    docs_groups = docs_to_clusters(docs, labels)  # dict where key is label ind, and value is list of docs values.
    for key, value in docs_groups.items():
        topic_names = tf_idf_top_words(value, words_count=5)
        tf_idf_labels[key] = create_topic_name(key, topic_names)
    return tf_idf_labels


def create_topic_names_using_tf_idf_with_optimization(docs, labels):
    # Optional: alternative function to 'create_topic_names_using_tf_idf'
    # Use this func to optimize TF-IDF selected words between topics
    porter = PorterStemmer()
    tf_idf_labels, topics_commonly, all_topics_words = dict(), dict(), dict()
    docs_groups = docs_to_clusters(docs, labels)
    for key, value in docs_groups.items():
        labels_array = tf_idf_top_words(value, words_count=7)
        all_topics_words[key] = labels_array
        for word in labels_array:
            stemmed_word = porter.stem(word)
            count = topics_commonly.get(stemmed_word, 0)
            topics_commonly[stemmed_word] = count + 1
    optimize_topic_words(all_topics_words, topics_commonly, tf_idf_labels, docs_groups, words_count=7,
                         final_words_count=3)
    return tf_idf_labels


def optimize_topic_words(all_topics_words, topics_commonly, tf_idf_labels, docs_groups, words_count=7,
                         final_words_count=5, tf_idf_common_threshold=0.4):
    porter = PorterStemmer()
    topics_amount = len(docs_groups.keys())
    for key, value in docs_groups.items():
        final_words = [""] * final_words_count
        replaced_count = 0
        for i in range(final_words_count - 1, -1, -1):
            stemmed_word = porter.stem(all_topics_words[key][i])
            if topics_commonly[stemmed_word] / topics_amount > tf_idf_common_threshold and replaced_count < words_count - final_words_count:
                alternative_word = all_topics_words[key][final_words_count + replaced_count]
                alternative_word_stemmed = porter.stem(alternative_word)
                if topics_commonly[stemmed_word] > topics_commonly[alternative_word_stemmed]:
                    print("stemmed word {} - replaced with {}".format(stemmed_word, alternative_word))
                    final_words[i] = alternative_word
                    replaced_count += 1
                else:
                    # The word is common but the alternative is even worst.
                    final_words[i] = all_topics_words[key][i]
                # IF we would like to extract this word from *some* of the topics, it should be done here.
                # topics_commonly[stemmed_word] = topics_commonly[stemmed_word] - 1
            else:
                final_words[i] = all_topics_words[key][i]
        tf_idf_labels[key] = create_topic_name(key, final_words)


# def create_division(ground_truth_df, random_percent):
#     # Helper function for create_random_divisions.
#     # Creates a division given ground truth division and random-error-percent
#     labels = ground_truth_df["label"].to_list()
#     result = labels.copy()  # [0,0,1,1,2,2]
#     labels_indexes = range(len(result))  # [0,1,2,3,4,5]
#     labels_set = sorted(set(labels))  # [0,1,2]
#     indexes_to_be_replaced = random.sample(labels_indexes, int((random_percent / 100) * len(labels_indexes)))
#     for index_to_be_replace in indexes_to_be_replaced:
#         alternative_labels = list(labels_set)
#         alternative_labels.remove(result[index_to_be_replace])
#         result[index_to_be_replace] = random.sample(alternative_labels, 1)[0]
#     return result


def extract_topics_labels(topics_names_dict):
    """
    returns list of topic names where each topic.
    each element returned is list of words that represent the topic.
    """
    topics_names = []
    for topic_ind in sorted(topics_names_dict.keys()):
        topics_names.append(topics_names_dict[topic_ind].split('_')[1:])
    return topics_names

def input_path_to_df(path):
    return pd.read_csv(path, sep='\t', usecols=['label', 'text'])


def detailed_topics_print(topics_names):
    print("\n", "*" * 5, "\tTopics\t", "*" * 5)
    for i, topic_name in enumerate(topics_names):
        print("Topic {} value is {}".format(i, topic_name))


def create_future_dirs():
    if not os.path.exists(".\\tmp_coh_exports"):
        os.mkdir(".\\tmp_coh_exports")
        os.mkdir(".\\tmp_coh_exports\\mnli")
        os.mkdir(".\\tmp_coh_exports\\topics_map")

def delete_tmp_dirs():
    if os.path.exists(".\\tmp_coh_exports"):
        shutil.rmtree(".\\tmp_coh_exports")


from tqdm import tqdm
from transformers import pipeline
import numpy as np
import pandas as pd
import random

from cohesion.utils import get_division


class ZeroShotClassifier:
    def __init__(self, path_dic_topic_id_to_name, path_df_docs_topics, topics, ground_truth_docs_list):
        self.model_name = "facebook/bart-large-mnli"
        self.df_docs_topics = pd.read_pickle(path_df_docs_topics)
        self.dic_topic_id_to_name = self.create_topic_ind_name_from_pkl(path_dic_topic_id_to_name)
        self.classifier = pipeline("zero-shot-classification",
                                   model=self.model_name)  # remove device=0 in case you don't use GPU.
        self.ground_truth_docs_list = ground_truth_docs_list
        self.docs, self.series_topics = self.get_topics_and_docs_filtered()

    def run_classify(self, division_labels, path_to_save, docs_percent, topics_percent):
        """
          Cohesion "Heart" - Uses MNLI to find similarity between docs and topics.
          results saved into 'path_to_save'
        """
        topic_names = []
        for i in sorted(self.dic_topic_id_to_name.keys()):
            topic_names.append(self.dic_topic_id_to_name[i])
        division_dict = get_division(division_labels)
        doc_values_to_test, docs_ind_of_selected_docs, topics_ind_of_selected_docs = self.get_selected_docs(
            division_dict, docs_percent)
        results = self.zero_shot_to_every_doc_and_random_topics(doc_values_to_test, topic_names,
                                                                topics_ind_of_selected_docs, docs_percent, topics_percent)
        df = self.create_df_zero_shot_results(topic_names, results, docs_ind_of_selected_docs)
        self.save_mnli_scores(df, path_to_save)

    def get_topics_and_docs_filtered(self):
        """
        drops docs that their topic id is not defined in dic_topic_id_to_name
        series topics is topic name for every doc that pass the filter
        returns docs and topics
        """
        series_topics = self.df_docs_topics["topic_ind"].apply(
            lambda x: self.dic_topic_id_to_name.get(x, np.nan)).dropna()
        docs = self.df_docs_topics["content"][series_topics.index].to_list()
        return docs, series_topics

    def get_selected_topics(self, topic_index_of_doc, topics, set_indexes_real_topics, topics_dict, topics_percent):
        selected_topics = []
        optional_topics = sorted(set_indexes_real_topics)
        optional_topics.remove(topic_index_of_doc)
        amount_of_topics_to_test = 1 + int(
            len(optional_topics) * topics_percent)  # % of relevant topics + 1 (to ensure negative topic as well) + 1 related topic! *MINIMUM amount of topics is 3(!)
        amount_of_topics_to_test = min(amount_of_topics_to_test, len(optional_topics))
        selected_topics_ind = list(random.sample(optional_topics, amount_of_topics_to_test))
        selected_topics_ind = [topic_index_of_doc] + selected_topics_ind  # related label is always at [0].
        for top_ind in selected_topics_ind:
            selected_topics.append(topics_dict[top_ind].split())
        return selected_topics

    def zero_shot_to_every_doc_and_random_topics(self, docs, topics, indexes_real_topic, docs_percent, topics_percent):
        topics_dict = dict()
        sorted_set_indexes_real_topic = sorted(set(indexes_real_topic))
        for i, top_ind in enumerate(sorted_set_indexes_real_topic):
            topics_dict[top_ind] = topics[i]
        results = []
        print_single = True
        set_indexes_real_topics = set(indexes_real_topic)
        for i, doc in enumerate(tqdm(docs)):
            topic_index_of_doc = indexes_real_topic[i]
            selected_topics = self.get_selected_topics(topic_index_of_doc, topics, set_indexes_real_topics, topics_dict, topics_percent)
            doc_topics_result = self.classifier(doc, selected_topics, multi_label=True)
            doc_topics_result["labels"] = [" ".join(label) for label in doc_topics_result["labels"]]
            results.append(doc_topics_result)
        return results

    def save_mnli_scores(self, df, path):
        df.to_pickle(path + ".pkl", protocol=4)
        df.to_csv(path + ".csv")

    def create_topic_ind_name_from_pkl(self, path):
        df_temp = pd.read_pickle(path)
        df_temp = df_temp[df_temp['topic_ind'] != -1]  # drops unassigned docs
        return dict(zip(df_temp["topic_ind"].to_list(), df_temp["topic_name"].to_list()))

    def get_selected_docs(self, division_dict, docs_percent):
        topics_ind_of_selected_docs = []
        docs_ind_of_selected_docs = []
        for i in division_dict.keys():
            docs_in_topic = division_dict.get(i)
            related_docs_amount = int(len(docs_in_topic) * docs_percent)
            related_docs_amount = related_docs_amount if related_docs_amount > 2 else min(2, len(docs_in_topic))
            try:
                selected_docs_from_this_topic = random.sample(docs_in_topic, related_docs_amount)
            except:
                print(related_docs_amount - 1)
                print(docs_in_topic)
            docs_ind_of_selected_docs += selected_docs_from_this_topic
            topics_ind_of_selected_docs += [i] * len(selected_docs_from_this_topic)
        doc_values_to_test = []
        for ind in docs_ind_of_selected_docs:
            doc_values_to_test.append(self.ground_truth_docs_list[ind])
        return doc_values_to_test, docs_ind_of_selected_docs, topics_ind_of_selected_docs

    def create_df_zero_shot_results(self, topic_names, results, docs_ind_of_selected_docs):
        topics_by_doc = self.series_topics.to_list()  # list of topics values, for every doc
        topic_names_with_ind = [str(i) + "_" + topic_name for i, topic_name in enumerate(topic_names)]
        df = pd.DataFrame(columns=["doc_num", "doc", "real_topic"] + topic_names_with_ind)
        for i, result in enumerate(results):
            doc_ind = docs_ind_of_selected_docs[i]
            doc_labels, doc_scores, doc_value = result["labels"], result["scores"], result["sequence"]
            doc_dict = {"doc_num": doc_ind, "doc": doc_value, "real_topic": topics_by_doc[doc_ind]}
            for topic_i, spec_topic in enumerate(self.dic_topic_id_to_name.values()):
                doc_topic_score = 0  # default value for doc - topic. there is no mandatory to test each topic against each doc.
                if spec_topic in doc_labels:
                    index_of_spec_topic = doc_labels.index(spec_topic)
                    doc_topic_score = doc_scores[index_of_spec_topic]
                doc_dict[str(topic_i) + "_" + spec_topic] = doc_topic_score
            df = df.append(doc_dict, ignore_index=True)
        return df


import pandas as pd

def cohesion_calculate(path):
    docs_col, topic_col = generate_docs_topics_collections(path)
    topics_with_scores = calculate_topics_score(docs_col, topic_col)
    score = calculate_topics_cohesion_ratio(topics_with_scores)
    return score

def generate_docs_topics_collections(path):
    data, topics = read_zero_shot_results(path)
    docs_collection = insert_data_to_docs_collection(data, topics)
    topics_collection = insert_data_to_topics_collection(topics)
    return docs_collection, topics_collection

def read_zero_shot_results(path):
    authorized_columns = ['Unnamed: 0', 'doc', 'real_topic']
    data = pd.read_csv(path)
    topics = data.columns
    topics_without_underscore = []
    for topic in topics:
        if topic in authorized_columns:
            topics_without_underscore.append(topic)
        elif topic == 'doc_num':
            continue
        else:
            split_topic = topic.split('_')
            topics_without_underscore.append([topic, split_topic[1]] if len(split_topic) == 2 else [topic, topic])
    return data, topics_without_underscore

# -------------- Insertion --------------
def insert_data_to_docs_collection(docs, topics):
    not_allow_topics = ['Unnamed: 0', 'doc', 'real_topic']
    list_to_insert = []
    scores = {}
    for index, row in docs.iterrows():
        for topic in topics:
            if topic not in not_allow_topics:
                scores[topic[1]] = row[topic[0]]
        object_doc = {'index': row['Unnamed: 0'], 'doc': row['doc'], 'real_topic': row['real_topic'], 'scores': scores.copy()}
        list_to_insert.append(object_doc)
        scores.clear()
    return list_to_insert

def insert_data_to_topics_collection(topics):
    not_allow_topics = ['Unnamed: 0', 'doc', 'real_topic']
    topics_dict = {}
    for index, topic in enumerate(topics):
        if topic in not_allow_topics:
            continue
        object_score = {'index': index - 3, 'description': topic[1], 'positive': 0, 'negative': 0}
        topics_dict[topic[1]] = object_score
    return topics_dict

# -------------- Calculations --------------

def calculate_topics_score(docs_collection, topics_collection):
    for index, (topic_name, topic_data) in enumerate(topics_collection.items()):
        positive, positive_length, negative, negative_length, positive_score, negative_score = 0, 0, 0, 0, 0, 0
        for doc in docs_collection:
            topic_doc_score = doc['scores'][topic_name]
            if doc['real_topic'] == topic_name:
                positive += topic_doc_score
                positive_length += 1
            elif topic_doc_score != 0:
                negative += topic_doc_score
                negative_length += 1
        positive_score = positive/positive_length
        negative_score = negative/negative_length
        topics_collection[topic_name]['positive'] = positive_score
        topics_collection[topic_name]['negative'] = negative_score
    return topics_collection

def calculate_topics_cohesion_ratio(topics_with_scores):
    # Main logic for Cohesion Formula aggregation
    positive, negative, total_score = 0, 0, 0
    for index, (topic_name, topic_data) in enumerate(topics_with_scores.items()):
      ratio = topic_data['positive']/topic_data['negative']
      if ratio>=1:
        topic_total_score = 1 - 1/(ratio+1)
      else:
        topic_total_score = ratio / (1+ratio)
      total_score += topic_total_score
    return total_score/len(topics_with_scores)


import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "cohesion_pipeline",
    version = "0.0.5",
    author = "Cohesion Team",
    author_email = "berdugogo@gmail.com",
    description = "cohesion_pipeline",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/Berdugo1994",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'nltk~=3.7',
        'scikit-learn~=1.1.1',
        'pandas~=1.3.0',
        'transformers~=4.20.1',
        'numpy~=1.21.0',
        'tqdm~=4.64.0',
    ],
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    python_requires = ">=3.8"
)